#include <windows.h>
#include <shellapi.h>

#include <string>
#include <vector>

#include "resource.h"
#define WM_FIXTRAY WM_USER + 400

HINSTANCE           hInst;
HWND                hwnd;
HMENU               hmenu, popup;
HICON               icon_large, icon_small;
std::string         app_name;
std::vector<HWND>   hidden_windows;
NOTIFYICONDATA      tray_notify;

std::string load_string(UINT uid, HINSTANCE hInst = NULL)
{
    char buffer[4096];
    LoadString(hInst, uid, buffer, 4096);
    return buffer;
}

void on_quit()
{
    for (std::vector<HWND>::size_type i = 0;
        i < hidden_windows.size(); ++i)
    {
        ShowWindowAsync(hidden_windows[i], SW_SHOW);
    }
    PostQuitMessage(0);
}

void on_help(HWND hwnd)
{
    MessageBox(hwnd, load_string(IDS_HELP_CONTENT, hInst).c_str(),
               app_name.c_str(), MB_ICONINFORMATION);
}

//
LRESULT CALLBACK WndProc(HWND hwnd, UINT uMessage,
                         WPARAM wparam, LPARAM lparam)
{
    switch(uMessage)
    {
    case WM_LBUTTONDOWN:
        {
            char buffer[4096];

            POINT point;
            GetCursorPos(&point);
            HWND handle = WindowFromPoint(point);
            HWND parent = GetParent(handle);

            while (parent != NULL)
            {
                handle = parent;
                parent = GetParent(handle);
            }

            if (GetWindowText(handle, buffer, 4096))
            {
                HMENU temp=GetSubMenu(popup, 1);
                hidden_windows.push_back(handle);
                AppendMenu(temp, MF_STRING,
                    ID_WINDOW_LIST + hidden_windows.size(), buffer);
                if (hidden_windows.size() == 1)
                    DeleteMenu(temp, ID_WINDOW_LIST, MF_BYCOMMAND);
                ShowWindowAsync(handle, SW_HIDE);
            }
        }
        break;
    case WM_CLOSE:
        on_quit();
        break;
    case WM_USER + 100:
        if(lparam == WM_LBUTTONUP || lparam == WM_RBUTTONUP)
        {
            POINT point;
            GetCursorPos(&point);
            SetForegroundWindow(hwnd);
            TrackPopupMenuEx(popup, TPM_LEFTALIGN | TPM_TOPALIGN,
                             point.x, point.y, hwnd, NULL);
            PostMessage(hwnd, WM_NULL, 0, 0);
        }
        break;
    case WM_COMMAND:
        {
            UINT item_id = LOWORD(wparam);
            if (item_id > ID_WINDOW_LIST)
            {
                ShowWindowAsync(hidden_windows[item_id - ID_WINDOW_LIST - 1],
                                SW_SHOW);
                HMENU temp=GetSubMenu(popup, 1);

                UINT id;
                for (id = item_id;
                    id <= ID_WINDOW_LIST + hidden_windows.size(); ++id)
                    DeleteMenu(temp, id, MF_BYCOMMAND);

                hidden_windows.erase
                    ((std::vector<HWND>::iterator)
                    &hidden_windows[item_id - ID_WINDOW_LIST - 1]);

                char buffer[4096];
                for (id = item_id;
                    id <= ID_WINDOW_LIST + hidden_windows.size(); ++id)
                {
                    if (GetWindowText(hidden_windows
                        [id - ID_WINDOW_LIST - 1], buffer, 4096))
                        AppendMenu(temp, MF_STRING, id, buffer);
                }

                if (hidden_windows.size() == 0)
                {
                    AppendMenu(temp, MF_STRING, ID_WINDOW_LIST,
                               load_string(IDS_NOTHING, hInst).c_str());
                    EnableMenuItem(temp, ID_WINDOW_LIST,
                                   MF_GRAYED|MF_BYCOMMAND);
                }
            }

            switch (item_id)
            {
            case ID_HIDE:
                SetCapture(hwnd);
                break;
            case ID_SHOWHELP:
                on_help(hwnd);
                break;
            case ID_EXIT:
                on_quit();
                break;
            default:
                return 0;
            }
        }
        break;
    case WM_FIXTRAY:
        Shell_NotifyIcon(NIM_DELETE, &tray_notify);
        Shell_NotifyIcon(NIM_ADD, &tray_notify);
        break;
    default:
        return DefWindowProc(hwnd, uMessage, wparam, lparam);
    }

    return 0;
}

BOOL InitApp(HINSTANCE hInstance)
{
    WNDCLASS wc;

    wc.style         = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc   = (WNDPROC) WndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance;
    wc.hIcon         = icon_large
                     = (HICON) LoadImage(hInstance,
                               MAKEINTRESOURCE(IDI_ICON), IMAGE_ICON,
                               32, 32, LR_DEFAULTCOLOR);
    icon_small       = (HICON) LoadImage(hInstance,
                               MAKEINTRESOURCE(IDI_ICON), IMAGE_ICON,
                               16, 16, LR_DEFAULTCOLOR);
    wc.hCursor       = LoadCursor(NULL,IDC_ARROW);
    wc.hbrBackground = (HBRUSH) (COLOR_BTNFACE + 1);
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = "TrayStor";

    if (!RegisterClass(&wc))
       return FALSE;
    return TRUE;
}

BOOL InitInst(HINSTANCE hInstance, int nCmdShow)
{
    hInst = hInstance;

    hwnd = CreateWindow("TrayStor",
                        NULL,
                        WS_OVERLAPPEDWINDOW,
                        CW_USEDEFAULT, 0,
                        CW_USEDEFAULT, 0,
                        NULL,
                        NULL,
                        hInstance,
                        NULL);

    if (!hwnd)
        return FALSE;

    return TRUE;
}

int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE,
                     LPSTR lpCmdLine, int nCmdShow)
{
    HANDLE hShared=CreateFileMapping((HANDLE) 0xffffffff, NULL,
        PAGE_READWRITE, 0, sizeof(HWND),
        "BabyFoxApplicationTrayWindowsStorage");
    HWND *pPrev=(HWND *)
        MapViewOfFile(hShared, FILE_MAP_ALL_ACCESS, 0, 0, 0);

    if (GetLastError() == ERROR_ALREADY_EXISTS)
    {
        SendMessage(*pPrev, WM_FIXTRAY, 0, 0);
        UnmapViewOfFile(pPrev);
        CloseHandle(hShared);
        return 0;
    }

    app_name = load_string(IDS_APPNAME, hInstance);
    MSG msg;

    if (!InitApp(hInstance))
    {
        UnmapViewOfFile(pPrev);
        CloseHandle(hShared);
        return 0xffffffff;
    }

    if (!InitInst(hInstance, nCmdShow))
    {
        UnmapViewOfFile(pPrev);
        CloseHandle(hShared);
        return 0xffffffff;
    }

    *pPrev=hwnd;
    UnmapViewOfFile(pPrev);

    hmenu=LoadMenu(hInst,MAKEINTRESOURCE(IDR_MENU));
    popup=GetSubMenu(hmenu, 0);

    lstrcpy(tray_notify.szTip, app_name.c_str());
    tray_notify.hIcon            = icon_small;
    tray_notify.uCallbackMessage = WM_USER + 100;
    tray_notify.uID              = 0;
    tray_notify.uFlags           = NIF_MESSAGE | NIF_ICON | NIF_TIP;
    tray_notify.hWnd             = hwnd;
    tray_notify.cbSize           = sizeof(NOTIFYICONDATA);

    Shell_NotifyIcon(NIM_ADD, &tray_notify);

    while (GetMessage(&msg, NULL, 0, 0))
    {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    DestroyMenu(popup);
    DestroyMenu(hmenu);
    Shell_NotifyIcon(NIM_DELETE, &tray_notify);
    DestroyIcon(icon_large);
    DestroyIcon(icon_small);
    CloseHandle(hShared);

    return msg.wParam;
}
